/******************************************************************************
 * 
 * CS105: Cachelab
 *
 * Starter code: eval.c
 *
 * General code for running and evaluating the cache simulation
 * 
 * You DO NOT need to modify this file. 
 *
 * (but you might find it helpful to set the macro MAX_LEN to a small value for
 * debugging purposes (i.e., so you can step through the whole thing in gdb 
 * without it taking forever).
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "caching.h"


#define MAX_LEN 65536 // I recommend changing this to 4 or 8 for debugging ~EB

int num_hits; // global to track num_hits
int num_misses; //global to track num_misses


/******************************************************************************
 *  access_direct is a wrapper function for accessing memory
 *  first checks whether a value is in cache. If so, reads from cache. If not,
 *  reads from memory and updates the cache
 *****************************************************************************/
int access_direct(int* addr){
  // breakdown address into tag, index, offset
  long tag; 
  byte index;
  byte offset; 
  parse_addr((void*) addr, &tag, &index, &offset);

  if(is_in_direct_cache(tag, index, offset)){ // if cache hit
    num_hits++; 

    // return cached value
    int val_cached = lookup_int_in_direct_cache(tag, index, offset);
    return val_cached;

  } else { // cache miss 
    num_misses++;

    // lookup value from memory
    int val = *addr; 

    // update cache
    update_direct_cache(tag,index, offset, (void*) addr);
    
    // return value
    return val;
  }  
}


/******************************************************************************
 * vector_multiplication multiplies two one-dimensional arrays of integers.
 * used by main function to evaluating caching
 *****************************************************************************/
long vector_multiplication(int* a, int* b, int len){
  long result = 0;
  for(int i = 0; i < len; i++){
    int x = access_direct(&(a[i]));
    int y = access_direct(&(b[i]));
    result += x*y;
  }
  return result;
}


/******************************************************************************
 * helper function for initializing simulation, allocates & inits cache + vecs
 *****************************************************************************/
int* initialize_simulation(){
  // allocate and initialize direct-mapped "cache"
  direct_cache = malloc(NUM_CACHELINES * sizeof(cacheline_t));
  for(int i = 0; i < NUM_CACHELINES; i++){
    (direct_cache + i)->valid = 0; // all cachelines are initially invalid
  }

  // allocate and initialize space for vectors
  int max_size = MAX_LEN * sizeof(int) * 2;
  int* vector_memory = (int*) malloc(max_size);
  for(int i = 0; i < max_size/sizeof(int); i++){
    vector_memory[i] = i;
  }
  
  return vector_memory;
}


/******************************************************************************
 * helper function for simulation. resets cache to empty and counters to 0
 *****************************************************************************/
void reset_simulation(){
  // re-initialize cache by setting all lines to invalid
  for(int i = 0; i < NUM_CACHELINES; i++){
    (direct_cache + i)->valid = 0; // all cachelines are initially invalid
  }
  
  // reset counters
  num_hits = 0;
  num_misses = 0;
}


/******************************************************************************
 * main function initialized memory and cache and then evaluates the cache
 *****************************************************************************/
int main(int argc, char** argv){
  // initialize simulation
  int* vector_memory = initialize_simulation();
  
  // evaluate cache performance
  for(int len = 2; len < MAX_LEN; len*=2){
    // reset simulation
    reset_simulation();

    int* a = vector_memory; // a occupies first len*sizeof(int) bytes
    int* b = vector_memory+len; // b occupies next len*sizeof(int) bytes
    long ans = vector_multiplication(a, b, len);
    printf("Len: %d, Answer: %ld, Cache hits: %f, Cache misses: %f\n",len,ans,
	   (double)num_hits/(num_hits+num_misses),
	   (double)num_misses/(num_hits+num_misses)); 

  }
  
  // cleanup
  free(vector_memory);
  free(direct_cache);

  return 0;
}
